/*
 * MemberAbstraction.hpp
 *
 *  Created on: 11. 7. 2017
 *	  Author: Jan Travnicek
 */

#ifndef _MEMBER_ABSTRACTION_HPP_
#define _MEMBER_ABSTRACTION_HPP_

#include <alib/memory>

#include <abstraction/NaryOperationAbstraction.hpp>
#include <abstraction/ReferenceAbstraction.hpp>

#include <registry/Registry.h>

namespace abstraction {

template < class ObjectType, class ReturnType, class ... ParamTypes >
class MemberAbstraction : public NaryOperationAbstraction < ReturnType, ObjectType &, ParamTypes ... > {
	std::function < ReturnType ( typename std::remove_reference < ObjectType >::type *, ParamTypes ... ) > m_callback;

public:
	MemberAbstraction ( std::function < ReturnType ( typename std::remove_reference < ObjectType >::type *, ParamTypes ... ) > callback ) : m_callback ( std::move ( callback ) ) {
	}

	bool run ( ) override {
		ext::array < std::pair < std::shared_ptr < OperationAbstraction >, bool >, sizeof ... ( ParamTypes ) + 1 > params = this->getParams ( );

		std::shared_ptr < OperationAbstraction > reference = std::make_shared < ReferenceAbstraction < typename std::remove_reference < ObjectType >::type > > ( );
		if ( ! reference->attachInput ( std::get < 0 > ( this->getParams ( ) ).first, 0, std::get < 0 > ( this->getParams ( ) ).second, true ) )
			throw std::invalid_argument ( "Can't bind the object of call to member." );

		if ( ! reference->eval ( ) )
			throw std::invalid_argument ( "Eval of object of call to member falsed." );

		std::get < 0 > ( params ) = std::make_pair ( reference, false );

		this->template run_helper < typename std::remove_reference < ObjectType >::type *, ParamTypes ... > ( m_callback, params );
		return true;
	}

};

} /* namespace abstraction */

#endif /* _MEMBER_ABSTRACTION_HPP_ */
